/* **********************************************************************
    Copyright 2003 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.rpi.sss.util.xml;

import java.io.IOException;
import java.io.Writer;
import java.util.HashMap;
import java.util.Iterator;
import javax.servlet.http.HttpServletResponse;

/** Class to emit XML
 *
 * @author Mike Douglass  douglm@rpi.edu
 */
public class XmlEmit {
  private Writer wtr;
  private boolean mustEmitNS;

  /** We need to map the namespaces onto a set of reasonable abbreviations
   * for the generated xml. New set created each request
   */
  private HashMap nsMap;

  private int nsIndex;

  /** The following allow us to tidy up the output a little.
   */
  int indent;
  private String blank = "                                       " +
                                "                                       ";
  private int blankLen = blank.length();

  /** construct an object which will be used to collect namespace names
   * during the first phase and emit xml afetr startEmit is called.
   */
  public XmlEmit() {
    nsMap = new HashMap();
    nsIndex = 0;
  }

  /** Emit any headers and namespace declarations
   */
  public void startEmit(HttpServletResponse resp) throws IOException {
    startEmit(resp.getWriter());
  }

  public void startEmit(Writer wtr) throws IOException {
    this.wtr = wtr;
    mustEmitNS = true;

    writeHeader();
    newline();
  }

  public void openTag(QName tag) throws IOException {
    blanks();
    openTagSameLine(tag);
    newline();
    indent += 2;
  }

  public void openTagSameLine(QName tag) throws IOException {
    lb();
    tagname(tag);
    rb();
  }

  public void closeTag(QName tag) throws IOException {
    indent -= 2;
    if (indent < 0) {
      indent = 0;
    }
    blanks();
    closeTagSameLine(tag);
    newline();
  }

  public void closeTagSameLine(QName tag) throws IOException {
    lb();
    wtr.write("/");
    tagname(tag);
    rb();
  }

  public void emptyTag(QName tag) throws IOException {
    blanks();
    emptyTagSameLine(tag);
    newline();
  }

  public void emptyTagSameLine(QName tag) throws IOException {
    lb();
    tagname(tag);
    wtr.write("/");
    rb();
  }

  /** Create the sequence<br>
   *  <tag>val</tag>
   */
  public void property(QName tag, String val) throws IOException {
    blanks();
    openTagSameLine(tag);
    wtr.write(val);
    closeTagSameLine(tag);
    newline();
  }

  /** Create the sequence<br>
   *  <tag><tagVal></tag>
   */
  public void propertyTagVal(QName tag, QName tagVal) throws IOException {
    blanks();
    openTagSameLine(tag);
    emptyTagSameLine(tagVal);
    closeTagSameLine(tag);
    newline();
  }

  public void flush() throws IOException {
    wtr.flush();
  }

  public void addNs(String val) {
    if (nsMap.get(val) != null) {
      return;
    }

    nsMap.put(val, "ns" + nsIndex);
    nsIndex++;
  }

  public String getNsAbbrev(String ns) {
    return (String)nsMap.get(ns);
  }

  /** Write out the tag name, adding the ns abbreviation.
   * Also add the namespace declarations if this is the first tag
   */
  private void tagname(QName tag) throws IOException {
    String ns = tag.getNamespaceURI();

    if (ns != null) {
      String abbr = getNsAbbrev(ns);

      if (abbr != null) {
        wtr.write(abbr);
        wtr.write(":");
      }
    }

    wtr.write(tag.getLocalPart());

    if (mustEmitNS) {
      Iterator nss = nsMap.keySet().iterator();

      while (nss.hasNext()) {
        wtr.write(" xmlns");

        ns = (String)nss.next();
        String abbr = (String)nsMap.get(ns);

        if (abbr != null) {
          wtr.write(":");
          wtr.write(abbr);
        }

        wtr.write("=\"");
        wtr.write(ns);
        wtr.write("\"");
      }

      mustEmitNS = false;
    }
  }

  /** Write out the xml header
   */
  private void writeHeader() throws IOException {
    wtr.write("<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n");
  }

  private void blanks() throws IOException {
    if (indent >= blankLen) {
      wtr.write(blank);
    } else {
      wtr.write(blank.substring(0, indent));
    }
  }

  private void lb() throws IOException {
    wtr.write("<");
  }

  private void rb() throws IOException {
    wtr.write(">");
  }

  private void newline() throws IOException {
    wtr.write("\n");
  }
}

