package edu.washington.cac.calfacade.shared;

import java.io.Serializable;
import java.util.Set;

/** The location of an <code>Event</code>
 *
 *  @version 1.0
 */
public class LocationVO implements Comparable, Serializable {
  private int id/* = CalFacadeDefs.unsavedItemKey*/;
  private String address, subaddress, link;

  /** The person who created the location */
  private UserVO creator;

  private boolean isPublic;

  /** Events referencing this location
   */
  protected Set events;

  public LocationVO() {
  }

  /** create a location with all fields specified
   *
   *  @param id          int unique id
   *  @param address     main address
   *  @param subaddress  secondary address
   *  @param link        String URL with more info
   *  @param creator     UserVO user that created it
   *  @param isPublic    true if location is viewable by everyone
   */
  public LocationVO(int id,
                    String address,
                    String subaddress,
                    String link,
                    UserVO creator,
                    boolean isPublic) {
    this.id = id;
    this.address = address;
    this.subaddress = subaddress;
    this.link = link;
    this.creator = creator;
    this.isPublic = isPublic;
  }

  public void setId(int val) {
    id = val;
  }

  /** Get the unique id of the location
   *
   *  @return the unique id of the location
   */
  public int getId() {
    return id;
  }

  public void setAddress(String val) {
    address = val;
  }

  /** Get the main address of the location
      @return the main address of the location
    */
  public String getAddress() {
    return address;
  }

  public void setSubaddress(String val) {
    subaddress = val;
  }

  /** Get the secondary address of the location
      @return the secondary address of the location
   */
  public String getSubaddress() {
    return subaddress;
  }

  /**
    Set the Location's URL
    @param link The new URL
   */
  public void setLink(String link) {
    this.link = link;
  }

  /** Get the link for the location
        @return the link for the location
      */
  public String getLink() {
    return link;
  }

  public void setCreator(UserVO val) {
    creator = val;
  }

  /**
    @return UserVO     creator of the location
    */
  public UserVO getCreator() {
    return creator;
  }

  public boolean getIsPublic() {
    return isPublic;
  }

  public void setIsPublic(boolean val) {
    isPublic = val;
  }

  public void setPublic(boolean val) {
    isPublic = val;
  }

  public boolean getPublic() {
    return isPublic;
  }

  /**
      Test whether the location is public
      @return whether the location is public
    */
  public boolean isPublic() {
    return isPublic;
  }

  /** Set the events - must be a set of EventVO
   *
   * @param val      Set of EventVO
   */
  public void setEvents(Set val) {
    events = val;
  }

  /** Get the events
   *
   * @param Set      of EventVO
   */
  public Set getEvents() {
    return events;
  }

  /* ====================================================================
   *                   Object methods
   * ==================================================================== */

  /**  Compares this object with the specified object for order.
   * Returns a negative integer, zero, or a positive integer as this object
   * is less than, equal to, or greater than the specified object.
   *
   */
  public int compareTo(Object o) {
    if (!(o instanceof LocationVO)) {
      return -1;
    }

    if (equals(o)) {
      return 0;
    }

    LocationVO that = (LocationVO)o;

    int res = CalFacadeUtil.cmpObjval(getAddress(), that.getAddress());

    if (res == 0) {
      res = CalFacadeUtil.cmpObjval(getSubaddress(), that.getSubaddress());
    }

    if (res == 0) {
      res = CalFacadeUtil.cmpObjval(getLink(), that.getLink());
    }

    if (res == 0) {
      res = CalFacadeUtil.cmpObjval(getCreator(), that.getCreator());
    }

    return res;
  }

  /** The id does not take part in these comparisons because we're presumably
   * doing this to see if two objects are the same without doing a db retrieval.
   * That is, one of the objects in a comparison may not have the id set.
   *
   * <p>This is particularly important with caching persistance engines such
   * as hibernate.
   *
   * <p>This is relevant for hashcode as the equals method requires that
   * hashcode return equal values for both objects in a comparison that are
   * to be considered equal.
   *
   * <p>As we have both objects to hand in the equals method we can check to
   * see if both ids are non-zero and use them in the comparison.
   */
  public int hashCode() {
    int hc = 1;

    if (getAddress() != null) {
      hc *= getAddress().hashCode();
    }

    if (getSubaddress() != null) {
      hc *= getSubaddress().hashCode();
    }

    if (getLink() != null) {
      hc *= getLink().hashCode();
    }

    if (getCreator() != null) {
      hc *= getCreator().hashCode();
    }

    if (getPublic()) {
      hc *= 2;
    }

    return hc;
  }

  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }

    if (!(obj instanceof LocationVO)) {
      return false;
    }

    LocationVO that = (LocationVO)obj;

    if (!CalFacadeUtil.eqObjval(getAddress(), that.getAddress())) {
      return false;
    }

    if (!CalFacadeUtil.eqObjval(getSubaddress(), that.getSubaddress())) {
      return false;
    }

    if (!CalFacadeUtil.eqObjval(getLink(), that.getLink())) {
      return false;
    }

    if (!CalFacadeUtil.eqObjval(getCreator(), that.getCreator())) {
      return false;
    }

    return getPublic() == that.getPublic();
  }

  public String toString() {
    StringBuffer sb = new StringBuffer();

    sb.append("LocationVO{id=");
    sb.append(id);
    sb.append(", address=");
    sb.append(address);
    sb.append(", subaddress=");
    sb.append(subaddress);
    sb.append(", link=");
    sb.append(link);
    sb.append(", creator=");
    UserVO c = getCreator();
    if (c == null) {
      sb.append("null");
    } else {
      sb.append(c.getId());
    }
    sb.append("}");

    return sb.toString();
  }

  public Object clone() {
    if (id <= CalFacadeDefs.maxReservedLocationId) {
      // It's a reserved immutable location. Just return it.
      return this;
    }

    return new LocationVO(id, address, subaddress, link,
                          (UserVO)creator.clone(), isPublic);
  }
}
