/* **********************************************************************
    Copyright 2005 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.rpi.cct.uwcal.caldav;

import edu.rpi.cct.uwcal.caldav.filter.Filter;
import edu.rpi.cct.webdav.servlet.common.Headers;
import edu.rpi.cct.webdav.servlet.common.MethodBase;
import edu.rpi.cct.webdav.servlet.common.PropFindMethod;
import edu.rpi.cct.webdav.servlet.common.WebdavMethods;
import edu.rpi.cct.webdav.servlet.shared.WebdavException;
import edu.rpi.cct.webdav.servlet.shared.WebdavNsIntf;
import edu.rpi.cct.webdav.servlet.shared.WebdavNsNode;
import edu.rpi.cct.webdav.servlet.shared.WebdavStatusCode;
import edu.rpi.cct.webdav.servlet.shared.WebdavTags;
import edu.rpi.sss.util.xml.QName;
import edu.rpi.sss.util.xml.XmlUtil;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import java.util.Iterator;
import java.util.Vector;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

/** Class called to handle OPTIONS. We should determine what the current
 * url refers to and send a response which shows the allowable methods on that
 * resource.
 *
 *   @author Mike Douglass   douglm@rpi.edu
 */
public class ReportMethod extends MethodBase {
  /* The parsed results go here. We see:
   *  1. Free-busy request
   *  2. Query - optional props + optional calendar data + filter
   *  3. Multi-get - optional props + optional calendar data + one or more hrefs
   */

  private FreeBusyQuery freeBusy;
  private PropFindMethod.PropRequest preq;
  private CalendarQueryCalendarData caldata;
  private Filter filter;
  private Vector hrefs;

  private final static int reportTypeQuery = 0;
  private final static int reportTypeMultiGet = 1;
  private final static int reportTypeFreeBusy = 2;
  private final static int reportTypeExpandProperty = 3;

  private int reportType;

  /** Called at each request
   */
  public void init() {
  }

  public void doMethod(HttpServletRequest req,
                        HttpServletResponse resp) throws WebdavException {
    if (debug) {
      trace("ReportMethod: doMethod");
    }

    Document doc = parseContent(req, resp);

    if (doc == null) {
      return;
    }

    startEmit(resp);

    int st = processDoc(doc);

    if (st != HttpServletResponse.SC_OK) {
      resp.setStatus(st);
      throw new WebdavException(st);
    }

    int depth = Headers.depth(req);

    if (debug) {
      trace("ReportMethod: depth=" + depth);
    }

    processResp(req, resp, depth);
  }

  /** ===================================================================
   *                   Private methods
   *  =================================================================== */

  /** We process the parsed document and produce a Collection of request
   * objects to process.
   */
  private int processDoc(Document doc) throws WebdavException {
    try {
      WebdavNsIntf intf = getNsIntf();

      Element root = doc.getDocumentElement();

      /* Get hold of the PROPFIND method instance - we need it to process
         possible prop requests.
       */
      PropFindMethod pm = (PropFindMethod)intf.findMethod(
            WebdavMethods.propFind);

      if (pm == null) {
        return HttpServletResponse.SC_INTERNAL_SERVER_ERROR;
      }

      if (nodeMatches(root, CaldavTags.calendarQuery)) {
        reportType = reportTypeQuery;
      } else if (nodeMatches(root, CaldavTags.calendarMultiget)) {
        reportType = reportTypeMultiGet;
      } else if (nodeMatches(root, CaldavTags.freeBusyQuery)) {
        reportType = reportTypeFreeBusy;
      } else if (nodeMatches(root, WebdavTags.expandProperty)) {
        reportType = reportTypeExpandProperty;
      } else {
        return HttpServletResponse.SC_BAD_REQUEST;
      }

      Element[] children = getChildren(root);

      for (int i = 0; i < children.length; i++) {
        Element curnode = children[i];

        String nodenm = curnode.getNodeName();
        /** The names come through looking something like A:allprop etc.
         */

        String nm = curnode.getLocalName();
        String ns = curnode.getNamespaceURI();

        if (debug) {
          trace("reqtype: " + nm + " ns: " + ns);
        }

        if (reportType == reportTypeFreeBusy) {
          /* We should only see time-range elements */
        } else if (reportType == reportTypeExpandProperty) {
        } else {
          /* Two possibilities:
               <!ELEMENT calendar-multiget
                            (DAV:allprop | DAV:propname | DAV:prop)?
                            calendar-data? DAV:href+>

               <!ELEMENT calendar-query
                       (DAV:allprop | DAV:propname | DAV:prop)?
                       calendar-data? filter>
           */

          /* First try for a property request */
          PropFindMethod.PropRequest pr = pm.tryPropRequest(curnode);

          if (pr != null) {
            if (preq != null) {
              if (debug) {
                trace("REPORT: preq not null");
              }
              return HttpServletResponse.SC_BAD_REQUEST;
            }
            preq = pr;
          } else if (nodeMatches(curnode, CaldavTags.calendarData)) {
            if (caldata != null) {
              if (debug) {
                trace("REPORT: caldata not null");
              }
              return HttpServletResponse.SC_BAD_REQUEST;
            }
            caldata = new CalendarQueryCalendarData(debug);
            int st = caldata.parse(curnode);

            if (st != HttpServletResponse.SC_OK) {
              return st;
            }
          } else if ((reportType == reportTypeQuery) &&
                     nodeMatches(curnode, CaldavTags.filter)) {
            if (filter != null) {
              if (debug) {
                trace("REPORT: filter not null");
              }
              return HttpServletResponse.SC_BAD_REQUEST;
            }
            filter = new Filter(debug);
            int st = filter.parse(curnode);

            if (st != HttpServletResponse.SC_OK) {
              return st;
            }
          } else if ((reportType == reportTypeMultiGet) &&
                     nodeMatches(curnode, WebdavTags.href)) {
            String href = XmlUtil.getElementContent(curnode);

            if ((href ==null) || (href.length() == 0)) {
              return HttpServletResponse.SC_BAD_REQUEST;
            }

            if (hrefs == null) {
              hrefs = new Vector();
            }

            hrefs.add(href);
          } else {
            if (debug) {
              trace("REPORT: unexpected element");
            }
            return HttpServletResponse.SC_BAD_REQUEST;
          }
        }
      }

      // Final validation
      if (reportType == reportTypeMultiGet) {
        if (hrefs == null) {
          // need at least 1
          return HttpServletResponse.SC_BAD_REQUEST;
        }
      } else if (reportType == reportTypeQuery) {
        if (filter == null) {
          // filter required
          return HttpServletResponse.SC_BAD_REQUEST;
        }
      }

      if (debug) {
        trace("REPORT: ");
        if (reportType == reportTypeFreeBusy) {
          trace("free-busy");
          freeBusy.dump();
        } else if (reportType == reportTypeQuery) {
          // Query - optional props + optional calendar data + filter
          if (caldata != null) {
            caldata.dump();
          } else {
            trace("No caldata");
          }

          filter.dump();
        } else if (reportType == reportTypeMultiGet) {
          // Multi-get - optional props + optional calendar data + one or more hrefs
          if (caldata != null) {
            caldata.dump();
          } else {
            trace("No caldata");
          }

          Iterator it = hrefs.iterator();
          while (it.hasNext()) {
            String href = (String)it.next();
            trace("    <DAV:href>" + href + "</DAV:href>");
          }
        } else {
        }
      }

      return HttpServletResponse.SC_OK;
    } catch (WebdavException wde) {
      return HttpServletResponse.SC_BAD_REQUEST;
    } catch (Throwable t) {
      System.err.println(t.getMessage());
      if (debug) {
        t.printStackTrace();
      }

      return HttpServletResponse.SC_INTERNAL_SERVER_ERROR;
    }
  }

  public void processResp(HttpServletRequest req,
                          HttpServletResponse resp,
                          int depth) throws WebdavException {
    resp.setStatus(WebdavStatusCode.SC_MULTI_STATUS);
    resp.setContentType("text/xml; charset=UTF-8");

    String resourceUri = getResourceUri(req);
    if (debug) {
      trace("About to get node at " + resourceUri);
    }

    WebdavNsNode node = getNsIntf().getNode(resourceUri);

    openTag(WebdavTags.multistatus);

    doNodeAndChildren(node, 0, depth);

    closeTag(WebdavTags.multistatus);

    flush();
  }

  /** Apply a node to a parsed request - or the other way - whatever.
   */
  private void doNodeAndChildren(WebdavNsNode node,
                                 int curDepth,
                                 int maxDepth) throws WebdavException {
    if (reportType == reportTypeFreeBusy) {
    } else {
      PropFindMethod pm = (PropFindMethod)getNsIntf().findMethod(
              WebdavMethods.propFind);

      pm.doNodeProperties(node, preq);
      if (reportType == reportTypeQuery) {
        // Do caldata + filter query
      } else {
        // Do caldata + hreq
      }
    }

    flush();

    curDepth++;

    if (curDepth > maxDepth) {
      return;
    }

    Iterator children = getNsIntf().getChildren(node);

    while (children.hasNext()) {
      WebdavNsNode child = (WebdavNsNode)children.next();

      doNodeAndChildren(child, curDepth, maxDepth);
    }
  }
}

