/* **********************************************************************
    Copyright 2005 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.rpi.cct.uwcal.caldav.filter;

import edu.rpi.cct.uwcal.caldav.CaldavTags;
import edu.rpi.cct.webdav.servlet.common.MethodBase;
import edu.rpi.cct.webdav.servlet.shared.WebdavException;
import edu.rpi.sss.util.xml.XmlUtil;

import java.text.ParseException;
import java.util.Date;
import javax.servlet.http.HttpServletResponse;

import net.fortuna.ical4j.util.DateFormat;

import org.apache.log4j.Logger;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

/** Class to represent a calendar-query filter
 *  <pre>
 *   <!ELEMENT filter comp-filter>
 *
 *   10.3.1 CALDAV:comp-filter XML Element
 *
 *    Name:
 *        comp-filter
 *    Namespace:
 *        urn:ietf:params:xml:ns:caldav
 *    Purpose:
 *        Limits the search to only the chosen component types.
 *    Description:
 *        The "name" attribute is an iCalendar component type (e.g., "VEVENT").
 *        When this element is present, the server should only return a component
 *        if it matches the filter, which is to say:
 *
 *    ("no is-defined element" OR "is-defined matches") AND
 *    ("no time-range element" OR "time-range matches") AND
 *    ("no sub-component filter" OR "all sub-component filters match") AND
 *    ("no property filter elements" OR "all property filters match")
 *
 *
 *    <!ELEMENT comp-filter (is-defined | time-range)?
 *                          comp-filter* prop-filter*>
 *
 *    <!ATTLIST comp-filter name CDATA #REQUIRED>
 *
 *    10.3.2 CALDAV:prop-filter XML Element
 *
 *    Name:
 *        prop-filter
 *    Namespace:
 *        urn:ietf:params:xml:ns:caldav
 *    Purpose:
 *        Limits the search to specific properties.
 *    Description:
 *        The "name" attribute MUST contain an iCalendar property name (e.g., "ATTENDEE"). When the 'prop-filter' executes, a property matches if:
 *
 *    ("no is-defined element" OR "is-defined matches") AND
 *    ("no time-range element" OR "time-range matches") AND
 *    ("no text match element" OR "text-match matches") AND
 *    ("no parameter filter elements" OR "all parameter filters match")
 *
 *    <!ELEMENT prop-filter (is-defined | time-range | text-match)?
 *                            param-filter*>
 *
 *    <!ATTLIST prop-filter name CDATA #REQUIRED>
 *
 *    10.3.3 CALDAV:param-filter XML Element
 *
 *    Name:
 *        param-filter
 *    Namespace:
 *        urn:ietf:params:xml:ns:caldav
 *    Purpose:
 *        Limits the search to specific parameters.
 *    Description:
 *        The "param-filter" element limits the search result to the set of resources containing properties with parameters that meet the parameter filter rules. When this filter executes, a parameter matches if:
 *
 *    ("is-defined matches" OR "text-match matches")
 *
 *    <!ELEMENT param-filter (is-defined | text-match) >
 *
 *    <!ATTLIST param-filter name CDATA #REQUIRED>
 *
 * </pre>
 *
 *   @author Mike Douglass   douglm@rpi.edu
 */
public class Filter {
  private boolean debug;

  protected transient Logger log;

  /** The root element.
   */
  private CompFilter filter;

  /* Set by internal methods to indicate failure reason.
   */
  private int status;

  public Filter(boolean debug) {
    this.debug = debug;
  }

  /** The given node must be the Filter element
   *
   * @return int     htp status
   */
  public int parse(Node nd) throws WebdavException {
    /* We expect exactly one comp-filter child. */

    Element[] children = getChildren(nd);

    if ((children.length != 1) ||
        (!MethodBase.nodeMatches(children[0], CaldavTags.compFilter))) {
      if (debug) {
        trace("<filter>: child count error: " + children.length);
      }
      return HttpServletResponse.SC_BAD_REQUEST;
    }

    filter = parseCompFilter((Node)children[0]);

    if (filter == null) {
      return status;
    }

    return HttpServletResponse.SC_OK;
  }

  /** The given node must be a comp-filter element
   *    <!ELEMENT comp-filter (is-defined | time-range)?
   *                          comp-filter* prop-filter*>
   *
   *    <!ATTLIST comp-filter name CDATA #REQUIRED>
   */
  private CompFilter parseCompFilter(Node nd) throws WebdavException {
    String name = getOnlyAttrVal(nd, "name");
    if (name == null) {
      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }

    CompFilter cf = new CompFilter(name);

    Element[] children = getChildren(nd);

    for (int i = 0; i < children.length; i++) {
      Node curnode = children[i];

      if (debug) {
        trace("compFilter element: " +
              curnode.getNamespaceURI() + " " +
              curnode.getLocalName());
      }

      if (MethodBase.nodeMatches(curnode, CaldavTags.isDefined)) {
        if (cf.getTimeRange() != null) {
          status = HttpServletResponse.SC_BAD_REQUEST;
          return null;
        }

        cf.setIsDefined(true);
      } else if (MethodBase.nodeMatches(curnode, CaldavTags.timeRange)) {
        if (cf.getIsDefined()) {
          status = HttpServletResponse.SC_BAD_REQUEST;
          return null;
        }

        cf.setTimeRange(parseTimeRange(curnode));

        if (cf.getTimeRange() == null) {
          return null;
        }
      } else if (MethodBase.nodeMatches(curnode, CaldavTags.compFilter)) {
        CompFilter chcf = parseCompFilter(curnode);

        if (chcf == null) {
          return null;
        }

        cf.addCompFilter(chcf);
      } else if (MethodBase.nodeMatches(curnode, CaldavTags.propFilter)) {
        PropFilter chpf = parsePropFilter(curnode);

        if (chpf == null) {
          return null;
        }

        cf.addPropFilter(chpf);
      } else {
        status = HttpServletResponse.SC_BAD_REQUEST;
        return null;
      }
    }

    return cf;
  }

  /** The given node must be a time-range element
   *  <!ELEMENT time-range EMPTY>
   *
   *  <!ATTLIST time-range start CDATA
   *                       end CDATA>
   *
   * e.g.        <C:time-range start="20040902T000000Z"
   *                           end="20040902T235959Z"/>
   *
   */
  private TimeRange parseTimeRange(Node nd) throws WebdavException {
    Date start = null;
    Date end = null;

    NamedNodeMap nnm = nd.getAttributes();

    /* draft 5 has neither attribute required - the intent is that either
       may be absent */

    if ((nnm == null) || (nnm.getLength() == 0)) {
      // Infinite time-range?
      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }

    int attrCt = nnm.getLength();

    try {
      DateFormat df = DateFormat.getInstance();
      Node nmAttr = nnm.getNamedItem("start");

      if (nmAttr != null) {
        attrCt--;
        start = df.parse(nmAttr.getNodeValue());
      }

      nmAttr = nnm.getNamedItem("end");

      if (nmAttr != null) {
        attrCt--;
        end = df.parse(nmAttr.getNodeValue());
      }
    } catch (ParseException pe) {
      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }

    if (attrCt != 0) {
      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }

    return new TimeRange(start, end);
  }

  /** The given node must be a prop-filter element
   *    <!ELEMENT prop-filter (is-defined | time-range | text-match)?
   *                            param-filter*>
   *
   *    <!ATTLIST prop-filter name CDATA #REQUIRED>
   */
  private PropFilter parsePropFilter(Node nd) throws WebdavException {
    String name = getOnlyAttrVal(nd, "name");
    if (name == null) {
      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }

    PropFilter pf = new PropFilter(name);

    Element[] children = getChildren(nd);
    boolean idTrTm = false; // flag is-defined | time-range | text-match

    for (int i = 0; i < children.length; i++) {
      Node curnode = children[i];

      if (debug) {
        trace("propFilter element: " +
              curnode.getNamespaceURI() + " " +
              curnode.getLocalName());
      }

      if (idTrTm) {
        // Only have param-filter*
        if (MethodBase.nodeMatches(curnode, CaldavTags.paramFilter)) {
          ParamFilter parf = parseParamFilter(curnode);

          if (parf == null) {
            return null;
          }

          pf.addParamFilter(parf);
        } else {
          status = HttpServletResponse.SC_BAD_REQUEST;
          return null;
        }
      } else {
        idTrTm = true;

        // one of is-defined | time-range | text-match
        if (MethodBase.nodeMatches(curnode, CaldavTags.isDefined)) {
          pf.setIsDefined(true);
        } else if (MethodBase.nodeMatches(curnode, CaldavTags.timeRange)) {
          pf.setTimeRange(parseTimeRange(curnode));

          if (pf.getTimeRange() == null) {
            return null;
          }
        } else if (MethodBase.nodeMatches(curnode, CaldavTags.textMatch)) {
          pf.setMatch(parseTextMatch(curnode));

          if (pf.getMatch() == null) {
            return null;
          }
        } else {
          status = HttpServletResponse.SC_BAD_REQUEST;
          return null;
        }
      }
    }

    return pf;
  }

  /** The given node must be a param-filter element
   *    <!ELEMENT param-filter (is-defined | text-match) >
   *
   *    <!ATTLIST param-filter name CDATA #REQUIRED>
   */
  private ParamFilter parseParamFilter(Node nd) throws WebdavException {
    String name = getOnlyAttrVal(nd, "name");
    if (name == null) {
      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }

    // Only one child - either is-defined | text-match
    Element child = getOnlyChild(nd);

    if (debug) {
      trace("paramFilter element: " +
            child.getNamespaceURI() + " " +
            child.getLocalName());
    }

    if (MethodBase.nodeMatches(child, CaldavTags.isDefined)) {
      return new ParamFilter(name, true);
    }

    if (MethodBase.nodeMatches(child, CaldavTags.textMatch)) {
      TextMatch match = parseTextMatch(child);

      if (match == null) {
        return null;
      }

      return new ParamFilter(name, match);
    }

    status = HttpServletResponse.SC_BAD_REQUEST;
    return null;
  }

  /** The given node must be a text-match element
   *  <!ELEMENT text-match #PCDATA>
   *
   *  <!ATTLIST text-match caseless (yes|no)>
   */
  private TextMatch parseTextMatch(Node nd) throws WebdavException {
    int numAttrs = XmlUtil.numAttrs(nd);

    Boolean caseless = null;
    if (numAttrs == 1) {
      caseless = yesNoAttr(nd, "caseless");
    }

    try {
      return new TextMatch(caseless, XmlUtil.getReqOneNodeVal(nd));
    } catch (Throwable t) {
      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }
  }

  private boolean absent(String val) {
    return (val == null) || (val.length() == 0);
  }

  private Element[] getChildren(Node nd) {
    try {
      return XmlUtil.getElementsArray(nd);
    } catch (Throwable t) {
      if (debug) {
        getLogger().error("<filter>: parse exception: ", t);
      }

      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }
  }

  private Element getOnlyChild(Node nd) {
    try {
      return XmlUtil.getOnlyElement(nd);
    } catch (Throwable t) {
      if (debug) {
        getLogger().error("<filter>: parse exception: ", t);
      }

      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }
  }

  /** Fetch required attribute. Return null for error
   */
  public String getOnlyAttrVal(Node nd, String name) {
    NamedNodeMap nnm = nd.getAttributes();

    if ((nnm == null) || (nnm.getLength() != 1)) {
      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }

    String res = XmlUtil.getAttrVal(nnm, name);
    if (res == null) {
      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }

    return res;
  }

  /** Fetch required yes/no Return null for error
   */
  public Boolean yesNoAttr(Node nd, String name) {
    NamedNodeMap nnm = nd.getAttributes();

    if ((nnm == null) || (nnm.getLength() == 0)) {
      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }

    Boolean res = null;

    try {
      res = XmlUtil.getYesNoAttrVal(nnm, name);
    } catch (Throwable t) {}

    if (res == null) {
      status = HttpServletResponse.SC_BAD_REQUEST;
      return null;
    }

    return res;
  }

  /** ===================================================================
   *                   Dump methods
   *  =================================================================== */

  public void dump() {
    trace("  <filter>");
    filter.dump(getLogger(), "    ");
    trace("  </filter>");
  }

  /** ===================================================================
   *                   Logging methods
   *  =================================================================== */

  protected Logger getLogger() {
    if (log == null) {
      log = Logger.getLogger(this.getClass());
    }

    return log;
  }

  protected void debugMsg(String msg) {
    getLogger().debug(msg);
  }

  protected void logIt(String msg) {
    getLogger().info(msg);
  }

  protected void trace(String msg) {
    getLogger().debug(msg);
  }
}

