/* **********************************************************************
    Copyright 2005 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.rpi.cct.webdav.servlet.common;

import edu.rpi.cct.webdav.servlet.shared.WebdavException;
import edu.rpi.cct.webdav.servlet.shared.WebdavNsIntf;
import edu.rpi.cct.webdav.servlet.shared.WebdavNsNode;
import edu.rpi.sss.util.xml.QName;
import edu.rpi.sss.util.xml.XmlEmit;
import edu.rpi.sss.util.xml.XmlUtil;

import java.io.InputStream;
import java.io.StringReader;
import java.net.URLDecoder;
import java.sql.Timestamp;
import java.text.SimpleDateFormat;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.Vector;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.ServletConfig;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/** Base class for all webdav servlet methods.
 */
public abstract class MethodBase {
  protected boolean debug;

  protected boolean dumpContent;

  protected transient Logger log;

//  private ServletConfig config;

  /** namespace interface for this request
   */
  protected WebdavNsIntf nsIntf;

  private String urlPrefix;

  private String resourceUri;

  private String content;

  protected XmlEmit xml;

  /** Called at each request
   */
  public abstract void init();

  public abstract void doMethod(HttpServletRequest req,
                                HttpServletResponse resp)
        throws WebdavException;

  /** Called at each request
   */
  public void init(WebdavNsIntf nsIntf,
                   boolean debug,
                   boolean dumpContent) throws WebdavException{
    this.nsIntf = nsIntf;
    this.debug = debug;
    this.dumpContent = dumpContent;

//    config = servlet.getServletConfig();
    urlPrefix = WebdavUtils.getUrlPrefix(nsIntf.getRequest());
    xml = nsIntf.getXmlEmit();

    content = null;
    resourceUri = null;

    init();
  }
/*
  public ServletConfig getConfig() {
    return config;
  }
*/
  public WebdavNsIntf getNsIntf() {
    return nsIntf;
  }

  public String getUrlPrefix() {
    return urlPrefix;
  }

  /** Get the decoded and fixed resource URI
   *
   * @param req      Servlet request object
   * @return String  fixed up uri
   */
  protected String getResourceUri(HttpServletRequest req)
      throws WebdavException{
    if (resourceUri != null) {
      return resourceUri;
    }

    String uri = req.getServletPath();

    if ((uri == null) || (uri.length() == 0)) {
      /* No path specified - set it to root. */
      uri = "/";
    }

    if (debug) {
      trace("uri: " + uri);
    }

    resourceUri = fixPath(uri);

    if (debug) {
      trace("resourceUri: " + resourceUri);
    }

    return resourceUri;
  }

  /** Get request content.
   *
   * @param req      Servlet request object
   * @return String  Body of the request
   */
  protected String getContent(HttpServletRequest req) throws WebdavException{
    if (content != null) {
      return content;
    }

    if (req.getContentLength() == 0) {
      return null;
    }

    try {
      String enc = req.getCharacterEncoding();
      if (enc == null) {
        enc = System.getProperty("file.encoding");
      }

      // Apparently we can get leading or trailing quotes
      if (enc.startsWith("\"")) {
        enc = enc.substring(1, enc.length());
      }

      if (enc.endsWith("\"")) {
        enc = enc.substring(0, enc.length() - 1);
      }

      int segLen;
      int totLen = 0;
      StringBuffer sb = new StringBuffer();
      InputStream is = req.getInputStream();

      byte[] seg = new byte[4096];  // IBM page size

      segLen = is.read(seg);
      while (segLen != -1) {
        String segStr = new String(seg, 0, segLen, enc);
        sb.append(segStr);

        segLen = is.read(seg);
      }

      content = sb.toString();

      if (dumpContent) {
        debugMsg(content);
      }

      return content;
    } catch (Throwable t) {
      throw new WebdavException(t);
    }
  }

  /** Parse thw Webdav request body, and return the DOM representation.
   *
   * @param req        Servlet request object
   * @param resp       Servlet response object for bad status
   * @return Document  Parsed body or null for no body
   * @exception WebdavException Some error occurred.
   */
  protected Document parseContent(HttpServletRequest req,
                                  HttpServletResponse resp)
      throws WebdavException{
    String cont = getContent(req);

    if (cont == null) {
      return null;
    }

    try {
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      factory.setNamespaceAware(true);

      DocumentBuilder builder = factory.newDocumentBuilder();
      return builder.parse(new InputSource(new StringReader(cont)));
    } catch (SAXException e) {
      resp.setStatus(HttpServletResponse.SC_BAD_REQUEST);
      throw new WebdavException(HttpServletResponse.SC_BAD_REQUEST);
    } catch (Throwable t) {
      resp.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
      throw new WebdavException(t);
    }
  }

  public static boolean nodeMatches(Node nd, QName tag) {
    String ns = nd.getNamespaceURI();
    if (ns == null) {
      if (tag.getNamespaceURI() != null) {
        return false;
      }
    } else if (!ns.equals(tag.getNamespaceURI())) {
      return false;
    }

    String ln = nd.getLocalName();
    if (ln == null) {
      if (tag.getLocalPart() != null) {
        return false;
      }
    } else if (!ln.equals(tag.getLocalPart())) {
      return false;
    }

    return true;
  }

  /** Return a path, beginning with a "/", after "." and ".." are removed.
   * If the parameter path attempts to go above the root we return null.
   *
   * @param path      String path to be fixed
   * @return String   fixed path
   **/
  protected String fixPath(String path) throws WebdavException {
    if (path == null) {
      return null;
    }

    String decoded;
    try {
      decoded = URLDecoder.decode(path, "UTF8");
    } catch (Throwable t) {
      throw new WebdavException(t);
    }

    if (decoded == null) {
      return (null);
    }

    /** Make any backslashes into forward slashes.
     */
    if (decoded.indexOf('\\') >= 0) {
      decoded = decoded.replace('\\', '/');
    }

    /** Ensure a leading '/'
     */
    if (!decoded.startsWith("/")) {
      decoded = "/" + decoded;
    }

    /** Remove all instances of '//'.
     */
    while (decoded.indexOf("//") >= 0) {
      decoded = decoded.replaceAll("//", "/");
    }

    if (decoded.indexOf("/.") < 0) {
      return decoded;
    }

    /** Somewhere we may have /./ or /../
     */

    StringTokenizer st = new StringTokenizer(decoded, "/");

    Vector v = new Vector();
    while (st.hasMoreTokens()) {
      String s = st.nextToken();

      if (s.equals(".")) {
        // ignore
      } else if (s.equals("..")) {
        // Back up 1
        if (v.size() == 0) {
          // back too far
          return null;
        }

        v.removeElementAt(v.size() - 1);
      } else {
        v.addElement(s);
      }
    }

    /** Reconstruct */
    StringBuffer sb = new StringBuffer();
    Enumeration e = v.elements();
    while (e.hasMoreElements()) {
      sb.append('/');
      sb.append((String)e.nextElement());
    }

    return sb.toString();
  }

  protected String formatISO8601Date(Timestamp val) {
    if (val == null) {
      return null;
    }

    String tm = val.toString();

    int i = tm.indexOf(" ");
    StringBuffer sb = new StringBuffer(tm.substring(0, i));
    sb.append("T");
    sb.append(tm.substring(i + 1));
    sb.append("Z");

    return sb.toString();
  }

  protected String formatHTTPDate(Timestamp val) {
    if (val == null) {
      return null;
    }

    return new SimpleDateFormat("E, dd MMM yyyy HH:mm:ss ").format(val) + "GMT";
  }

  /** Entity tags are defined in RFC2068 - they are supposed to provide some
   * sort of indication the data has changed - e.g. a checksum.
   * <p>There are weak and strong tags
   */
  protected String getEntityTag(WebdavNsNode node, boolean strong)
      throws WebdavException {
    String val = getNsIntf().getEtagValue(node, strong);

    if (strong) {
      return "\"" + val + "\"";
    }

    return "W/\"" + val + "\"";
  }

  /** ===================================================================
   *                   XmlUtil wrappers
   *  =================================================================== */

  protected Element[] getChildren(Node nd) throws WebdavException {
    try {
      return XmlUtil.getElementsArray(nd);
    } catch (Throwable t) {
      if (debug) {
        getLogger().error(this, t);
      }

      throw new WebdavException(t);
    }
  }

  protected Element getOnlyChild(Node nd) throws WebdavException {
    try {
      return XmlUtil.getOnlyElement(nd);
    } catch (Throwable t) {
      if (debug) {
        getLogger().error(this, t);
      }

      throw new WebdavException(t);
    }
  }

  /** ===================================================================
   *                   XmlEmit wrappers
   *  =================================================================== */

  protected void startEmit(HttpServletResponse resp) throws WebdavException {
    try {
      xml.startEmit(resp);
    } catch (Throwable t) {
      throw new WebdavException(t);
    }
  }

  public void addNs(String val) {
    xml.addNs(val);
  }

  public String getNsAbbrev(String ns) {
    return xml.getNsAbbrev(ns);
  }

  protected void openTag(QName tag) throws WebdavException {
    try {
      xml.openTag(tag);
    } catch (Throwable t) {
      throw new WebdavException(t);
    }
  }

  protected void closeTag(QName tag) throws WebdavException {
    try {
      xml.closeTag(tag);
    } catch (Throwable t) {
      throw new WebdavException(t);
    }
  }

  public void emptyTag(QName tag) throws WebdavException {
    try {
      xml.closeTag(tag);
    } catch (Throwable t) {
      throw new WebdavException(t);
    }
  }

  public void property(QName tag, String val) throws WebdavException {
    try {
      xml.property(tag, val);
    } catch (Throwable t) {
      throw new WebdavException(t);
    }
  }

  public void propertyTagVal(QName tag, QName tagVal) throws WebdavException {
    try {
      xml.propertyTagVal(tag, tagVal);
    } catch (Throwable t) {
      throw new WebdavException(t);
    }
  }

  protected void flush() throws WebdavException {
    try {
      xml.flush();
    } catch (Throwable t) {
      throw new WebdavException(t);
    }
  }

  /** ===================================================================
   *                   Logging methods
   *  =================================================================== */

  protected Logger getLogger() {
    if (log == null) {
      log = Logger.getLogger(this.getClass());
    }

    return log;
  }

  protected void debugMsg(String msg) {
    getLogger().debug(msg);
  }

  protected void logIt(String msg) {
    getLogger().info(msg);
  }

  protected void trace(String msg) {
    getLogger().debug(msg);
  }
}

