/* **********************************************************************
    Copyright 2005 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.rpi.cct.webdav.servlet.common;

import edu.rpi.cct.webdav.servlet.shared.WebdavException;
import edu.rpi.cct.webdav.servlet.shared.WebdavNsIntf;
import edu.rpi.sss.util.servlets.ServletUtil;
import edu.rpi.sss.util.servlets.io.CharArrayWrappedResponse;
import edu.rpi.sss.util.xml.XmlEmit;

import java.io.CharArrayWriter;
import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.PrintStream;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpServletResponseWrapper;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;

import org.apache.log4j.Logger;

/** WebDAV Servlet.
 * This abstract servlet handles the request/response nonsense and calls
 * abstract routines to interact with an underlying data source.
 *
 * @author Mike Douglass   douglm@rpi.edu
 * @version 1.0
 */
public abstract class WebdavServlet extends HttpServlet {
  protected boolean debug;

  protected boolean dumpContent;

  protected transient Logger log;

  protected HashMap methods;

  protected Properties props;

  protected boolean anonymous;

  /** Some sort of identifying string for logging
   */
  public abstract String getId();

  public abstract WebdavNsIntf getNsIntf(HttpServletRequest req,
                                         ServletConfig config,
                                         Properties props)
      throws WebdavException;

  protected void service(HttpServletRequest req,
                         HttpServletResponse resp)
      throws ServletException, IOException {
    WebdavNsIntf intf = null;

    try {
      debug = !"0".equals(getInitParameter("debug"));
      if (debug) {
        debugMsg("entry: " + req.getMethod());
        dumpRequest(req);
      }

      anonymous = req.getRemoteUser() == null;
      XmlEmit xml = new XmlEmit();

      intf = getNsIntf(req, getServletConfig(), props);

      intf.init(this, req, props, xml);

      initMethods(intf, debug, dumpContent);

      if (dumpContent) {
        resp = new CharArrayWrappedResponse(resp,
                                            getLogger(), debug);
      }

      resp.setStatus(HttpServletResponse.SC_OK);

      String methodName = req.getMethod();

      MethodBase method = (MethodBase)methods.get(methodName.toUpperCase());

      if (method == null) {
        logIt("No method for '" + methodName + "'");

        // ================================================================
        //     Set the correct response
        // ================================================================
      } else {
        method.doMethod(req, resp);
      }
    } catch (WebdavException wde) {
      getLogger().error(this, wde);
      resp.sendError(wde.getStatusCode());
    } catch (Throwable t) {
      getLogger().error(this, t);
      resp.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
    } finally {
      if (intf != null) {
        intf.close();
      }

      if (dumpContent) {
        CharArrayWrappedResponse wresp = (CharArrayWrappedResponse)resp;
        String str = wresp.toString();

        debugMsg("------------------------ Dump of response -------------------");
        debugMsg(str);
        debugMsg("---------------------- End dump of response -----------------");

        resp = (HttpServletResponse)wresp.getResponse();
        resp.setContentLength(str.length());
        resp.getWriter().write(str);
      }
    }
  }

  public void initMethods(WebdavNsIntf nsIntf,
                          boolean debug,
                          boolean dumpContent) throws WebdavException{
    addMethods();

    Iterator mnames = getMethodNames().iterator();
    while (mnames.hasNext()) {
      MethodBase mb = (MethodBase)methods.get(mnames.next());

      if (mb != null) {
        mb.init(nsIntf, debug, dumpContent);
      }
    }
  }

  public void addMethods() throws WebdavException{
    methods = new HashMap();

    methods.put("ACL", new AclMethod());
    methods.put("COPY", new CopyMethod());
    methods.put("GET", new GetMethod());
    methods.put("HEAD", new HeadMethod());
    methods.put("OPTIONS", new OptionsMethod());
    methods.put("PROPFIND", new PropFindMethod());

    if (!anonymous) {
      methods.put("DELETE", new DeleteMethod());
      methods.put("MKCOL", new MkcolMethod());
      methods.put("MOVE", new MoveMethod());
      methods.put("POST", new PostMethod());
      methods.put("PROPPATCH", new PropPatchMethod());
      methods.put("PUT", new PutMethod());

      methods.put("LOCK", new LockMethod());
      methods.put("UNLOCK", new UnlockMethod());
    }
  }

  /**
   *
   */
  public void init(ServletConfig config) throws ServletException {
    super.init(config);

    dumpContent = "true".equals(config.getInitParameter("dumpContent"));

    getResources(config);
  }

  public Collection getMethodNames() {
    return methods.keySet();
  }

  public HashMap getMethods() {
    return methods;
  }

  public Properties getProperties() {
    return props;
  }

  private void getResources(ServletConfig config) throws ServletException {
    String resname = config.getInitParameter("application");

    if (resname != null) {
      InputStream is;

      ClassLoader classLoader =
          Thread.currentThread().getContextClassLoader();
      if (classLoader == null) {
        classLoader = this.getClass().getClassLoader();
      }
      is = classLoader.getResourceAsStream(resname + ".properties");

      props = new Properties();
      try {
        props.load(is);
      } catch (IOException ie) {
        throw new ServletException(ie);
      }
    }
  }

  public void dumpRequest(HttpServletRequest req) {
    Logger log = getLogger();

    try {
      Enumeration names = req.getHeaderNames();

      String title = "Request headers";

      log.debug(title);

      while (names.hasMoreElements()) {
        String key = (String)names.nextElement();
        String val = req.getHeader(key);
        log.debug("  " + key + " = \"" + val + "\"");
      }

      names = req.getParameterNames();

      title = "Request parameters";

      log.debug(title + " - global info and uris");
      log.debug("getRequestURI = " + req.getRequestURI());
      log.debug("getRemoteUser = " + req.getRemoteUser());
      log.debug("getRequestedSessionId = " + req.getRequestedSessionId());
      log.debug("HttpUtils.getRequestURL(req) = " + req.getRequestURL());
      log.debug("query=" + req.getQueryString());
      log.debug("contentlen=" + req.getContentLength());
      log.debug("request=" + req);
      log.debug("parameters:");

      log.debug(title);

      while (names.hasMoreElements()) {
        String key = (String)names.nextElement();
        String val = req.getParameter(key);
        log.debug("  " + key + " = \"" + val + "\"");
      }
    } catch (Throwable t) {
    }
  }

  public Logger getLogger() {
    if (log == null) {
      log = Logger.getLogger(this.getClass());
    }

    return log;
  }

  public void debugMsg(String msg) {
    getLogger().debug(msg);
  }

  public void logIt(String msg) {
    getLogger().info(msg);
  }
}
