/* **********************************************************************
    Copyright 2005 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.rpi.cct.webdav.servlet.shared;

import edu.rpi.sss.util.xml.XmlEmit;
import edu.rpi.cct.webdav.servlet.common.MethodBase;
import edu.rpi.cct.webdav.servlet.common.WebdavServlet;

import java.io.Reader;
import java.io.Serializable;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Properties;
import javax.servlet.http.HttpServletRequest;

import org.apache.log4j.Logger;

/** This acts as an interface to the underlying namespace for which this
 * servlet is acting as a gateway. This could be a file system, a set of
 * dynamically created objects or some sort of CMS for example.
 *
 * A namespace consists of a number of nodes which may be containers for
 * other nodes or leaf nodes.
 *
 * All nodes are considered identical in their capabilities, that is, a
 * non-terminal node might contain content.
 *
 * Some nodes are aliases of other nodes (e.g. symlinks in a unix file
 * system). By default these aliases will be followed.
 *
 *   @author Mike Douglass   douglm@rpi.edu
 */
public abstract class WebdavNsIntf implements Serializable {
  protected boolean debug;

  protected transient Logger log;

  protected WebdavServlet servlet;

  protected Properties props;

  private HttpServletRequest req;

  protected XmlEmit xml;

  /** Called before any other method is called to allow initialisation to
   * take place at the first or subsequent requests
   */
  public void init(WebdavServlet servlet,
                   HttpServletRequest req,
                   Properties props,
                   XmlEmit xml) throws WebdavIntfException {
    this.servlet = servlet;
    this.req = req;
    this.props = props;
    this.xml = xml;

    debug = "9".equals(props.get("edu.rpi.sss.debug"));

    xml.addNs(WebdavTags.namespace);
    addNamespace();
  }

  public XmlEmit getXmlEmit() {
    return xml;
  }

  public HttpServletRequest getRequest() {
    return req;
  }

  public WebdavServlet getServlet() {
    return servlet;
  }

  public MethodBase findMethod(String name) {
    return (MethodBase)getServlet().getMethods().get(name);
  }

  /** Called on the way out to allow resources to be freed.
   */
  public abstract void close();

  /** Returns the supported locks for the supportedlock property.
   *
   * <p>To ensure these will work always provide the full namespace "DAV:"
   * for example, the result for supported exclusive and shared write locks
   * would be the string
   *
   *  "&lt;DAV:lockentry&gt;" +
   *  "  &lt;DAV:lockscope&gt;&lt;DAV:exclusive/&gt;&lt;DAV:/lockscope&gt;" +
   *  "  &lt;DAV:locktype&gt;&lt;DAV:write/&gt;&lt;DAV:/locktype&gt;" +
   *  "&lt;DAV:/lockentry&gt;" +
   *  "&lt;DAV:lockentry&gt;" +
   *  "  &lt;DAV:lockscope&gt;&lt;DAV:shared/&gt;&lt;DAV:/lockscope&gt;" +
   *  "&lt;DAV:/lockentry&gt;"
   */
  public abstract String getSupportedLocks();

  /** Returns true if the namespace supports access control
   */
  public abstract boolean getAccessControl();

  /** Add any namespaces for xml tag names in requests and responses.
   * An abbreviation will be supplied by the servlet.

   * The name should be globally unique in a global sense so don't return
   * something like "RPI:"
   *
   * <p>Something more like "http://ahost.rpi.edu/webdav/"
   */
  public abstract void addNamespace() throws WebdavIntfException;

  /** Return the complete URL describing the location of the object
   * represented by the node
   *
   * @param node             node in question
   */
  public abstract String getLocation(WebdavNsNode node)
      throws WebdavIntfException;

  /** Retrieves a node by URI, following any links.
   *
   * @param uri              String URI of the node to retrieve
   * @return WebdavNsNode    node specified by the URI or the node aliased by
   *                         the node at the URI.
   */
  public abstract WebdavNsNode getNode(String uri)
      throws WebdavIntfException;

  /** Retrieves a node by URI optionally following the last alias.
   *
   * @param uri              String URI of the node to retrieve
   * @param followAlias      boolean true is we should retrieve the target
   *                         of the alias if the last element is an alias.
   *                         otherwise the node itself is returned
   * @return WebdavNsNode    node specified by the URI or the node aliased by
   *                         the node at the URI if followAlias is true.
   */
  public abstract WebdavNsNode getNode(String uri, boolean followAlias)
      throws WebdavIntfException;

  /** Stores/updates an object.
   *
   * @param node             node in question
   */
  public abstract void putNode(WebdavNsNode node)
      throws WebdavIntfException;

  /** Deletes a node from the namespace.
   *
   * @param node             node in question
   */
  public abstract void delete(WebdavNsNode node)
      throws WebdavIntfException;

  /** Returns the immediate children of a node.
   *
   * @param node             node in question
   * @return Iterator        over WebdavNsNode children
   */
  public abstract Iterator getChildren(WebdavNsNode node)
      throws WebdavIntfException;

  /** Returns the parent of a node.
   *
   * @param node             node in question
   * @return WebdavNsNode    node's parent, or null if the specified node
   *                         is the root
   */
  public abstract WebdavNsNode getParent(WebdavNsNode node)
      throws WebdavIntfException;

  /** Entity tags are defined in RFC2068 - they are supposed to provide some
   * sort of indication the data has changed - e.g. a checksum.
   * <p>There are weak and strong tags
   *
   * <p>This methods should return a suitable value for that tag.
   */
  public abstract String getEtagValue(WebdavNsNode node, boolean strong)
      throws WebdavIntfException;

  /** Returns all the namespace specific properties for the given node
   * Properties that can be supplied by using node values should not be
   * included.
   *
   * @param node             node in question
   * @return Enumeration     proeprties
   */
  public abstract Enumeration getProperties(WebdavNsNode node)
      throws WebdavIntfException;

  /** Returns an InputStream for the content.
   *
   * @param node             node in question
   * @return Reader          A reader for the content.
   */
  public abstract Reader getContent(WebdavNsNode node)
      throws WebdavIntfException;

  /** Set the content from a Reader
   *
   * @param node             node in question.
   * @param content          String body value.
   * @param create           true if this is a creation
   * @return WebdavNsNode    possibly modified node referring to new object
   */
  public abstract WebdavNsNode putContent(WebdavNsNode node, String content,
                                          boolean create)
      throws WebdavIntfException;

  /** Create a new node.
   *
   * @param node             node to create with new uri set
   */
  public abstract void create(WebdavNsNode node)
      throws WebdavIntfException;

  /** Creates an alias to another node.
   *
   * @param alias       alias node that should be created with uri and
   *                    targetUri set
   */
  public abstract void createAlias(WebdavNsNode alias)
      throws WebdavIntfException;

  /* ====================================================================
   *                Property value methods
   * ==================================================================== */

  /** Generate content type.
   *
   * @param xml       xml emitter class
   * @param node      for which we are generating resource type
   */
  public void generatePropContenttype(WebdavNsNode node)
          throws WebdavIntfException {
    String ct = node.getContentType();
    if (ct != null) {
      try {
        xml.property(WebdavTags.getcontenttype, ct);
      } catch (Throwable t) {
        throw new WebdavIntfException(t);
      }
    }
  }

  /** Generate default webdav resource type.
   *
   * @param xml       xml emitter class
   * @param node      for which we are generating resource type
   */
  public void generatePropResourcetype(WebdavNsNode node)
          throws WebdavIntfException {
    try {
      if (node.getCollection()) {
        xml.propertyTagVal(WebdavTags.resourcetype,
                           WebdavTags.collection);
      }
    } catch (Throwable t) {
      throw new WebdavIntfException(t);
    }
  }

  /* ====================================================================
   *                        Protected methods
   * ==================================================================== */

  protected Logger getLogger() {
    if (log == null) {
      log = Logger.getLogger(this.getClass());
    }

    return log;
  }

  protected void debugMsg(String msg) {
    getLogger().debug(msg);
  }

  protected void logIt(String msg) {
    getLogger().info(msg);
  }
}
