/* **********************************************************************
    Copyright 2005 Rensselaer Polytechnic Institute.

    All worldwide rights reserved. A license to use, copy, modify and
    distribute this software for noncommercial research purposes only is
    hereby granted, provided that this copyright notice and accompanying
    disclaimer is not modified or removed from the software.

    DISCLAIMER: The software is distributed" AS IS" without any express or
    implied warranty, including but not limited to, any implied warranties
    of merchantability or fitness for a particular purpose or any warrant)'
    of non-infringement of any current or pending patent rights. The authors
    of the software make no representations about the suitability of this
    software for any particular purpose. The entire risk as to the quality
    and performance of the software is with the user. Should the software
    prove defective, the user assumes the cost of all necessary servicing,
    repair or correction. In particular, neither Rensselaer Polytechnic
    Institute, nor the authors of the software are liable for any indirect,
    special, consequential, or incidental damages related to the software,
    to the maximum extent the law permits.
*/

package edu.rpi.cct.webdav.servlet.shared;

import java.io.Serializable;
import java.sql.Timestamp;

import org.apache.log4j.Logger;

/** Represents a node in the underlying namespace for which this
 * servlet is acting as a gateway. This could be a file system, a set of
 * dynamically created objects or some sort of CMS for example.
 *
 *   @author Mike Douglass   douglm@rpi.edu
 */
public abstract class WebdavNsNode implements Serializable {
  protected boolean debug;

  /** Does the resource exist? */
  protected boolean exists = true;

  private transient Logger log;

  /** Uri of the node. These are relative to the root of the namespace this
   * interface represents and should start with a "/". For example, if this
   * namespace is part of a file system starting at uabc in /var/local/uabc
   * and we are referring to a directory x at /var/local/uabc/a/x then the
   * uri should be /a/x/
   */
  protected String uri;

  /** Suitable for display
   */
  protected String name;

  protected Timestamp creDate;

  protected Timestamp lastmodDate;

  //  ????? locks??????

  /** True if this node is a collection
   */
  protected boolean collection;

  /** True if GET is allowed
   */
  protected boolean allowsGet;

  /** ...................................................................
   *                   Content fields
   * We assume these get filled in when we get the content.
   *  ................................................................... */

  protected String contentLang;

  protected int contentLen;

  protected String contentType;

  /** ...................................................................
   *                   Alias fields
   *  ................................................................... */

  /** true if this is an alias
   */
  protected boolean alias;

  protected String targetUri;

  public WebdavNsNode(boolean debug) {
    this.debug = debug;
  }

  /* ====================================================================
   *                   Abstract methods
   * ==================================================================== */

  public abstract void setOwner() throws WebdavIntfException;

  /** Should return a value suitable for insertion in the owner dav tag.
   */
  public abstract String getOwner() throws WebdavIntfException;

  /* ====================================================================
   *                   Property methods
   * ==================================================================== */

  /** This method is called before each setter/getter takes any action.
   * It allows the concrete implementation to defer some expensive
   * operation to just before the first call.
   *
   * @param content     boolean flag indicating if this is a content related
   *                    property - that is a property which requires fetching
   *                    and/or rendering the content
   */
  public void init(boolean content) throws WebdavIntfException {
  }

  public void setExists(boolean val) throws WebdavIntfException {
    exists = val;
  }

  public boolean getExists() throws WebdavIntfException {
    return exists;
  }

  public void setUri(String val) throws WebdavIntfException {
    init(false);
    uri = val;
  }

  public String getUri() throws WebdavIntfException {
    init(false);
    return uri;
  }

  public void setCreDate(Timestamp val) throws WebdavIntfException {
    init(false);
    creDate = val;
  }

  public Timestamp getCreDate() throws WebdavIntfException {
    init(false);
    return creDate;
  }

  public void setLastmodDate(Timestamp val) throws WebdavIntfException {
    init(false);
    lastmodDate = val;
  }

  public Timestamp getLastmodDate() throws WebdavIntfException {
    init(false);
    return lastmodDate;
  }

  public void setName(String val) throws WebdavIntfException {
    init(false);
    name = val;
  }

  public String getName() throws WebdavIntfException {
    init(false);
    return name;
  }

  public void setCollection(boolean val) throws WebdavIntfException {
    collection = val;
  }

  public boolean getCollection() throws WebdavIntfException {
    return collection;
  }

  public void setAllowsGet(boolean val) throws WebdavIntfException {
    allowsGet = val;
  }

  public boolean getAllowsGet() throws WebdavIntfException {
    return allowsGet;
  }

  public void setContentLang(String val) throws WebdavIntfException {
    init(false);
    contentLang = val;
  }

  public String getContentLang() throws WebdavIntfException {
    init(false);
    return contentLang;
  }

  public void setContentLen(int val) throws WebdavIntfException {
    init(true);
    contentLen = val;
  }

  public int getContentLen() throws WebdavIntfException {
    init(true);
    return contentLen;
  }

  public void setContentType(String val) throws WebdavIntfException {
    init(false);
    contentType = val;
  }

  /** A content type of null implies no content (or we don't know) */
  public String getContentType() throws WebdavIntfException {
    init(false);
    return contentType;
  }

  public void setAlias(boolean val) throws WebdavIntfException {
    init(false);
    alias = val;
  }

  public boolean getAlias() throws WebdavIntfException {
    init(false);
    return alias;
  }

  public void setTargetUri(String val) throws WebdavIntfException {
    init(false);
    targetUri = val;
  }

  public String getTargetUri() throws WebdavIntfException {
    init(false);
    return targetUri;
  }

  /* ********************************************************************
   *                        Protected methods
   * ******************************************************************** */

  protected Logger getLogger() {
    if (log == null) {
      log = Logger.getLogger(this.getClass());
    }

    return log;
  }

  protected void debugMsg(String msg) {
    getLogger().debug(msg);
  }

  protected void logIt(String msg) {
    getLogger().info(msg);
  }

  /* ********************************************************************
   *                        Object methods
   * ******************************************************************** */

  public int hashCode() {
    return uri.hashCode();
  }

  public boolean equals(Object o) {
    if (o == this) {
      return true;
    }

    if (!(o instanceof WebdavNsNode)) {
      return false;
    }

    WebdavNsNode that = (WebdavNsNode)o;

    return uri.equals(that.uri);
  }
}
